/*
 * Kernel interrupt demo module for DNP/2110-3V and other
 * Author  : MHA@SSV <info@dilnetpc.com> <m.hasewinkel@web.de>
 * Date    : 27.08.2004
 * Version : 1.00
 * License : GPL
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "ssvirq.h"

static unsigned long ssvirq_count = 0, *ssvirq_userptr = NULL;
static int ssvirq_used = 0, irqnr = SSVIRQ_IRQNUMBER;

#ifdef MODULE_AUTHOR
 MODULE_AUTHOR("MHA@SSV <info@dilnetpc.com> <m.hasewinkel@web.de>");
 MODULE_DESCRIPTION("Kernel interrupt demo module for DNP/2110-3V and other");
 MODULE_LICENSE("GPL");
#endif
/* possible module parameter - insmod ssvirq.o irqnr=X */
MODULE_PARM(irqnr, "i");       /* default on INT1 = IRQ2 for DNP/2110-3V */

/* irq handler */
void ssvirq_interrupt(int irq, void *dev_id, struct pt_regs *regs) {

    ssvirq_count++;

    if (ssvirq_userptr != NULL) {
	/* put to userspace program */
	put_user(ssvirq_count, ssvirq_userptr);
    }
}

/* ioctl method - ioctl(device, SSVIRQ_SETPOINTER, &irqcount)... */
static int ssvirq_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg) {

    int	returncode = 0;

	switch (cmd) {
		case SSVIRQ_GETVALUE:
		    /* put "ssvirq_count" immediately to userspace program */
			put_user(ssvirq_count, ((unsigned long *)arg));
			break;
		case SSVIRQ_SETPOINTER:
		    /* get pointer for "ssvirq_count" at userspace program */
			ssvirq_userptr = (unsigned long *)arg;
			break;
		default:
		    /* unknown IOCTL */
		        returncode = -EINVAL;
			break;
	}
    return returncode;
}

/* open method - device = open("/dev/ssvirq", O_RDWR)... */
static int ssvirq_open(struct inode * inode, struct file * file) {

    /* only one program could open the device */
    if (ssvirq_used != 0) {
	return -EBUSY;
    }
    MOD_INC_USE_COUNT; 
    ssvirq_used = 1;
    ssvirq_userptr = NULL;
    return 0;
}

/* close method - close(device)... */
static int ssvirq_release(struct inode * inode, struct file * file) {

    ssvirq_userptr = NULL;
    ssvirq_used = 0;
    MOD_DEC_USE_COUNT;
    return 0;
}

static struct file_operations ssvirq_fops = {
    ioctl:		ssvirq_ioctl,
    open:		ssvirq_open,
    release:		ssvirq_release,
};

/* load kernel module - insmod ssvirq.o... */
static int __init ssvirq_init(void) {

    /* register irq handler */
    if (request_irq(irqnr, ssvirq_interrupt, SA_INTERRUPT,
		    "ssvirq", NULL)) {
        printk(KERN_ERR "ssvirq: can't get assigned irq %i\n", irqnr);
	return -EIO;
    }

    /* register character device handler */
    if (register_chrdev(SSVIRQ_MAJOR,"ssvirq",&ssvirq_fops)) {
	printk(KERN_ERR "ssvirq: unable to get major %d\n", SSVIRQ_MAJOR);
	free_irq(irqnr, NULL);
	return -EIO;
    }
    printk(KERN_INFO "ssvirq module installed.\n");

    return 0;
}

/* unload kernel module - rmmod ssvirq... */
static void __exit ssvirq_cleanup(void) {

    /* unregister irq handler */
    free_irq(irqnr, NULL);
    /* unregister character device handler */
    unregister_chrdev(SSVIRQ_MAJOR, "ssvirq");
    printk(KERN_INFO "ssvirq module removed.\n");
}

EXPORT_NO_SYMBOLS;

module_init(ssvirq_init);
module_exit(ssvirq_cleanup);

